# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Fabric library for Rattail itself
"""

from __future__ import unicode_literals
from __future__ import absolute_import

from fabric.api import sudo, env, cd

from rattail.fablib import make_deploy, make_system_user, mkdir


deploy = make_deploy(__file__)


def bootstrap_rattail(home='/var/lib/rattail', uid=None, shell='/bin/bash'):
    """
    Bootstrap a basic Rattail software environment.
    """
    make_system_user('rattail', home=home, uid=uid, shell=shell)
    sudo('adduser {0} rattail'.format(env.user))

    mkdir('/etc/rattail')
    mkdir('/srv/rattail')
    mkdir('/var/log/rattail', owner='rattail:rattail', mode='0775')

    mkdir('/srv/rattail/init')
    deploy('daemon', '/srv/rattail/init/daemon')
    deploy('check-rattail-daemon', '/usr/local/bin/check-rattail-daemon')
    deploy('soffice', '/srv/rattail/init/soffice')
    # TODO: deprecate / remove these
    deploy('bouncer', '/srv/rattail/init/bouncer')
    deploy('datasync', '/srv/rattail/init/datasync')
    deploy('filemon', '/srv/rattail/init/filemon')


def upgrade_rattail_db(envname=None, envpath=None, config=None, user='rattail'):
    """
    Upgrade a Rattail database in the "default" supported way
    """
    if not (envname or envpath):
        raise ValueError("Must specify value for either 'envname' or 'envpath'")
    if not envpath:
        envpath = '/srv/envs/{}'.format(envname)
    if not config:
        config = 'app/rattail.conf'
    with cd(envpath):
        sudo('bin/alembic --config {} upgrade heads'.format(config), user=user)


def deploy_rattail_sudoers(remote_path='/etc/sudoers.d/rattail'):
    """
    Deploy the common sudoers file for rattail.
    """
    deploy.sudoers('sudoers', remote_path)



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Console Stuff
"""

from __future__ import unicode_literals, absolute_import

import sys

import progressbar


class Progress(object):
    """
    Provides a console-based progress bar.
    """

    def __init__(self, message, maximum, stdout=None):
        self.stdout = stdout or sys.stderr
        self.stdout.write("\n{}...({:,d} total)\n".format(message, maximum))
        widgets = [progressbar.Percentage(), ' ', progressbar.Bar(), ' ', progressbar.ETA()]
        self.progress = progressbar.ProgressBar(maxval=maximum, widgets=widgets).start()

    def update(self, value):
        self.progress.update(value)
        return True

    def destroy(self):
        self.stdout.write("\n")



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import os
from unittest import TestCase

from fixture import TempIO

from rattail.config import make_config
from rattail.filemon import config
from rattail.filemon import Action
from rattail.exceptions import ConfigurationError


class TestProfile(TestCase):

    def setUp(self):
        self.config = make_config([])
        self.config.set(u'rattail.filemon', u'foo.actions', u'bar')

    def test_empty_config_means_empty_profile(self):
        profile = config.Profile(self.config, u'nonexistent_key')
        self.assertEqual(len(profile.dirs), 0)
        self.assertFalse(profile.watch_locks)
        self.assertTrue(profile.process_existing)
        self.assertFalse(profile.stop_on_error)
        self.assertEqual(len(profile.actions), 0)

    def test_action_must_specify_callable(self):
        self.assertRaises(ConfigurationError, config.Profile, self.config, u'foo')

    def test_action_must_not_specify_both_func_and_class_callables(self):
        self.config.set(u'rattail.filemon', u'foo.action.bar.class', u'baz')
        self.config.set(u'rattail.filemon', u'foo.action.bar.func', u'baz')
        self.assertRaises(ConfigurationError, config.Profile, self.config, u'foo')

    def test_action_with_func_callable(self):
        self.config.set(u'rattail.filemon', u'foo.action.bar.func', u'os:remove')
        profile = config.Profile(self.config, u'foo')
        self.assertEqual(len(profile.actions), 1)
        action = profile.actions[0]
        self.assertEqual(action.spec, u'os:remove')
        self.assertTrue(action.action is os.remove)

    def test_action_with_class_callable(self):
        self.config.set(u'rattail.filemon', u'foo.action.bar.class', u'rattail.filemon:Action')
        profile = config.Profile(self.config, u'foo')
        self.assertEqual(len(profile.actions), 1)
        action = profile.actions[0]
        self.assertEqual(action.spec, u'rattail.filemon:Action')
        self.assertTrue(isinstance(action.action, Action))

    def test_action_with_args(self):
        self.config.set(u'rattail.filemon', u'foo.action.bar.func', u'shutil:move')
        self.config.set(u'rattail.filemon', u'foo.action.bar.args', u'/dev/null')
        profile = config.Profile(self.config, u'foo')
        self.assertEqual(len(profile.actions), 1)
        action = profile.actions[0]
        self.assertEqual(len(action.args), 1)
        self.assertEqual(action.args[0], u'/dev/null')

    def test_action_with_kwargs(self):
        self.config.set(u'rattail.filemon', u'foo.action.bar.func', u'rattail.filemon.actions:raise_exception')
        self.config.set(u'rattail.filemon', u'foo.action.bar.kwarg.message', u"Hello World")
        profile = config.Profile(self.config, u'foo')
        self.assertEqual(len(profile.actions), 1)
        action = profile.actions[0]
        self.assertEqual(len(action.kwargs), 1)
        self.assertEqual(action.kwargs[u'message'], u"Hello World")

    def test_action_with_default_retry(self):
        self.config.set(u'rattail.filemon', u'foo.action.bar.func', u'rattail.filemon.actions:noop')
        profile = config.Profile(self.config, u'foo')
        self.assertEqual(len(profile.actions), 1)
        action = profile.actions[0]
        self.assertEqual(action.retry_attempts, 1)
        self.assertEqual(action.retry_delay, 0)

    def test_action_with_valid_configured_retry(self):
        self.config.set(u'rattail.filemon', u'foo.action.bar.func', u'rattail.filemon.actions:noop')
        self.config.set(u'rattail.filemon', u'foo.action.bar.retry_attempts', u'42')
        self.config.set(u'rattail.filemon', u'foo.action.bar.retry_delay', u'100')
        profile = config.Profile(self.config, u'foo')
        self.assertEqual(len(profile.actions), 1)
        action = profile.actions[0]
        self.assertEqual(action.retry_attempts, 42)
        self.assertEqual(action.retry_delay, 100)

    def test_action_with_invalid_configured_retry(self):
        self.config.set(u'rattail.filemon', u'foo.action.bar.func', u'rattail.filemon.actions:noop')
        self.config.set(u'rattail.filemon', u'foo.action.bar.retry_attempts', u'-1')
        self.config.set(u'rattail.filemon', u'foo.action.bar.retry_delay', u'-1')
        profile = config.Profile(self.config, u'foo')
        self.assertEqual(len(profile.actions), 1)
        action = profile.actions[0]
        self.assertEqual(action.retry_attempts, 1)
        self.assertEqual(action.retry_delay, 0)

    def test_normalize_dirs(self):
        tmp = TempIO()
        dir1 = tmp.mkdir(u'dir1')
        # dir2 will be pruned due to its not existing
        dir2 = tmp.mkdir(u'dir2')
        os.rmdir(dir2)
        # file1 will be pruned due to its not being a directory
        file1 = tmp.putfile(u'file1', u'')
        self.config.set(u'rattail.filemon', u'foo.action.bar.func', u'os:remove')
        self.config.set(u'rattail.filemon', u'foo.dirs', u' '.join([u'"{0}"'.format(d) for d in [dir1, dir2, file1]]))
        profile = config.Profile(self.config, u'foo')
        self.assertEqual(len(profile.dirs), 1)
        self.assertEqual(profile.dirs[0], dir1)


class TestLoadProfiles(TestCase):

    def setUp(self):
        self.tmp = TempIO()
        self.config = make_config([])
        self.config.set(u'rattail.filemon', u'monitor', u'foo, bar')
        self.config.set(u'rattail.filemon', u'foo.dirs', u'"{0}"'.format(self.tmp))
        self.config.set(u'rattail.filemon', u'foo.actions', u'delete')
        self.config.set(u'rattail.filemon', u'foo.action.delete.func', u'os:remove')
        self.config.set(u'rattail.filemon', u'bar.dirs', u'"{0}"'.format(self.tmp))
        self.config.set(u'rattail.filemon', u'bar.actions', u'delete')
        self.config.set(u'rattail.filemon', u'bar.action.delete.func', u'os:remove')

    def test_returns_all_profiles_specified_in_monitor_option(self):
        monitored = config.load_profiles(self.config)
        self.assertEqual(len(monitored), 2)
        # leave profiles intact but replace monitor option with one key only
        self.config.set(u'rattail.filemon', u'monitor', u'foo')
        monitored = config.load_profiles(self.config)
        self.assertEqual(len(monitored), 1)

    def test_monitor_option_must_be_specified(self):
        # TODO: This seems hacky.
        self.config.parser.remove_option('rattail.filemon', 'monitor')
        self.assertRaises(ConfigurationError, config.load_profiles, self.config)

    def test_profiles_which_define_no_watched_folders_are_pruned(self):
        monitored = config.load_profiles(self.config)
        self.assertEqual(len(monitored), 2)
        # remove foo's watched folder(s)
        # TODO: This seems hacky.
        self.config.parser.remove_option('rattail.filemon', 'foo.dirs')
        monitored = config.load_profiles(self.config)
        self.assertEqual(len(monitored), 1)

    def test_profiles_which_define_no_actions_are_pruned(self):
        monitored = config.load_profiles(self.config)
        self.assertEqual(len(monitored), 2)
        # remove foo's actions
        # TODO: This seems hacky.
        self.config.parser.remove_option('rattail.filemon', 'foo.actions')
        monitored = config.load_profiles(self.config)
        self.assertEqual(len(monitored), 1)

    def test_fallback_to_legacy_mode(self):
        # replace 'monitor' option with 'monitored' and update profiles accordingly
        # TODO: This seems hacky.
        self.config.parser.remove_option('rattail.filemon', 'monitor')
        self.config.set(u'rattail.filemon', u'monitored', u'foo, bar')
        self.config.set(u'rattail.filemon', u'foo.dirs', u"['{0}']".format(self.tmp))
        self.config.set(u'rattail.filemon', u'foo.actions', u"['os:remove']")
        self.config.set(u'rattail.filemon', u'bar.dirs', u"['{0}']".format(self.tmp))
        self.config.set(u'rattail.filemon', u'bar.actions', u"['os:remove']")
        monitored = config.load_profiles(self.config)
        self.assertEqual(len(monitored), 2)
        profiles = list(monitored.values())
        self.assertTrue(isinstance(profiles[0], config.LegacyProfile))
        self.assertTrue(isinstance(profiles[1], config.LegacyProfile))


class TestLegacyProfile(TestCase):

    def setUp(self):
        self.config = make_config([])

    def test_empty_config_means_empty_profile(self):
        profile = config.LegacyProfile(self.config, u'nonexistent_key')
        self.assertEqual(len(profile.dirs), 0)
        self.assertFalse(profile.watch_locks)
        self.assertTrue(profile.process_existing)
        self.assertFalse(profile.stop_on_error)
        self.assertEqual(len(profile.actions), 0)

    def test_action_with_spec_only(self):
        self.config.set(u'rattail.filemon', u'foo.actions', u"['os:remove']")
        profile = config.LegacyProfile(self.config, u'foo')
        self.assertEqual(len(profile.actions), 1)
        spec, action, args, kwargs = profile.actions[0]
        self.assertEqual(spec, u'os:remove')
        self.assertTrue(action is os.remove)

    def test_action_with_spec_and_args(self):
        self.config.set(u'rattail.filemon', u'foo.actions', u"[('shutil:move', u'/dev/null')]")
        profile = config.LegacyProfile(self.config, u'foo')
        self.assertEqual(len(profile.actions), 1)
        spec, action, args, kwargs = profile.actions[0]
        self.assertEqual(spec, u'shutil:move')
        self.assertEqual(len(args), 1)
        self.assertEqual(args[0], u'/dev/null')

    def test_normalize_dirs(self):
        tmp = TempIO()
        dir1 = tmp.mkdir(u'dir1')
        # dir2 will be pruned due to its not existing
        dir2 = tmp.mkdir(u'dir2')
        os.rmdir(dir2)
        # file1 will be pruned due to its not being a directory
        file1 = tmp.putfile(u'file1', u'')
        self.config.set(u'rattail.filemon', u'foo.dirs', u"[{0}]".format(u', '.join([u"'{0}'".format(d) for d in [dir1, dir2, file1]])))
        profile = config.LegacyProfile(self.config, u'foo')
        self.assertEqual(len(profile.dirs), 1)
        self.assertEqual(profile.dirs[0], dir1)


class TestLoadLegacyProfiles(TestCase):

    def setUp(self):
        self.tmp = TempIO()
        self.config = make_config([])
        self.config.set(u'rattail.filemon', u'monitored', u'foo, bar')
        self.config.set(u'rattail.filemon', u'foo.dirs', u"['{0}']".format(self.tmp))
        self.config.set(u'rattail.filemon', u'foo.actions', u"['os:remove']")
        self.config.set(u'rattail.filemon', u'bar.dirs', u"['{0}']".format(self.tmp))
        self.config.set(u'rattail.filemon', u'bar.actions', u"['os:remove']")

    def test_returns_all_profiles_specified_in_monitor_option(self):
        monitored = config.load_legacy_profiles(self.config)
        self.assertEqual(len(monitored), 2)
        # leave profiles intact but replace monitored option with one key only
        self.config.set(u'rattail.filemon', u'monitored', u'foo')
        monitored = config.load_legacy_profiles(self.config)
        self.assertEqual(len(monitored), 1)

    def test_monitor_option_must_be_specified(self):
        # TODO: This seems hacky.
        self.config.parser.remove_option('rattail.filemon', 'monitored')
        self.assertRaises(ConfigurationError, config.load_legacy_profiles, self.config)

    def test_profiles_which_define_no_watched_folders_are_pruned(self):
        monitored = config.load_legacy_profiles(self.config)
        self.assertEqual(len(monitored), 2)
        # remove foo's watched folder(s)
        # TODO: This seems hacky.
        self.config.parser.remove_option('rattail.filemon', 'foo.dirs')
        monitored = config.load_legacy_profiles(self.config)
        self.assertEqual(len(monitored), 1)

    def test_profiles_which_define_no_actions_are_pruned(self):
        monitored = config.load_legacy_profiles(self.config)
        self.assertEqual(len(monitored), 2)
        # remove foo's actions
        # TODO: This seems hacky.
        self.config.parser.remove_option('rattail.filemon', 'foo.actions')
        monitored = config.load_legacy_profiles(self.config)
        self.assertEqual(len(monitored), 1)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals, absolute_import

import unittest

from rattail import exceptions


class TestRecipientsNotFound(unittest.TestCase):

    def test_init(self):
        self.assertRaises(TypeError, exceptions.RecipientsNotFound)
        exc = exceptions.RecipientsNotFound('testing')
        self.assertEqual(exc.key, 'testing')

    def test_str(self):
        exc = exceptions.RecipientsNotFound('testing')
        self.assertEqual(str(exc),
                         b"No recipients found in config for 'testing' emails.  Please set "
                         "'testing.to' (or 'default.to') in the [rattail.mail] section.")

    def test_unicode(self):
        exc = exceptions.RecipientsNotFound('testing')
        self.assertEqual(unicode(exc),
                         "No recipients found in config for 'testing' emails.  Please set "
                         "'testing.to' (or 'default.to') in the [rattail.mail] section.")



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Fabric library for Apache web server
"""

from __future__ import unicode_literals, absolute_import

import re

from fabric.api import sudo, abort

from rattail.fablib import apt


def install():
    """
    Install the Apache web service
    """
    apt.install('apache2')


def get_version():
    """
    Fetch the version of Apache running on the target system
    """
    result = sudo('apache2 -v')
    if result.succeeded:
        match = re.match(r'^Server version: Apache/(\d+\.\d+)\.\d+ \(.*\)', result)
        if match:
            return float(match.group(1))


def install_wsgi(python_home=None):
    """
    Install the mod_wsgi Apache module, with optional ``WSGIPythonHome`` value.
    """
    apt.install('libapache2-mod-wsgi')
    if python_home:
        if get_version() == 2.2:
            sudo('echo WSGIPythonHome {} > /etc/apache2/conf.d/wsgi'.format(python_home))
        else: # assuming 2.4
            sudo('echo WSGIPythonHome {} > /etc/apache2/conf-available/wsgi.conf'.format(python_home))
            enable_conf('wsgi')


def enable_conf(*names):
    """
    Enable the given Apache configurations
    """
    for name in names:
        sudo('a2enconf {}'.format(name))


def enable_mod(*names):
    """
    Enable the given Apache modules
    """
    for name in names:
        sudo('a2enmod {}'.format(name))


def enable_site(*names):
    """
    Enable the given Apache site(s)
    """
    for name in names:
        sudo('a2ensite {}'.format(name))


def deploy_conf(deployer, local_path, name, enable=False):
    """
    Deploy a config snippet file to Apache
    """
    version = get_version()
    if version == 2.2:
        deployer(local_path, '/etc/apache2/conf.d/{}.conf'.format(name))
    else:
        deployer(local_path, '/etc/apache2/conf-available/{}.conf'.format(name))
        if enable:
            enable_conf(name)


def deploy_site(deployer, local_path, name, enable=False, **kwargs):
    """
    Deploy a file to Apache sites-available
    """
    apache_version = get_version()
    if apache_version == 2.2:
        path = '/etc/apache2/sites-available/{}'.format(name)
    else:
        path = '/etc/apache2/sites-available/{}.conf'.format(
            '000-default' if name == 'default' else name)
    context = kwargs.pop('context', {})
    context['apache_version'] = apache_version
    deployer(local_path, path, context=context, **kwargs)
    if enable and name != 'default':
        enable_site(name)


def restart():
    """
    Restart the Apache web service
    """
    sudo('service apache2 restart')


def start():
    """
    Start the Apache web service
    """
    sudo('service apache2 start')


def stop():
    """
    Stop the Apache web service
    """
    sudo('service apache2 stop')



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Windows registry utils
"""

from __future__ import unicode_literals, absolute_import


def RegDeleteTree(key, subkey):
    """
    This is a clone of ``win32api.RegDeleteTree()``, since that apparently
    requires Vista or later.
    """
    import pywintypes
    import win32api
    import win32con
    import winerror

    def delete_contents(key):
        subkeys = []
        for name, reserved, class_, mtime in win32api.RegEnumKeyEx(key):
            subkeys.append(name)
        for subkey_name in subkeys:
            subkey = win32api.RegOpenKeyEx(key, subkey_name, 0, win32con.KEY_ALL_ACCESS)
            delete_contents(subkey)
            win32api.RegCloseKey(subkey)
            win32api.RegDeleteKey(key, subkey_name)
        values = []
        i = 0
        while True:
            try:
                name, value, type_ = win32api.RegEnumValue(key, i)
            except pywintypes.error, e:
                if e[0] == winerror.ERROR_NO_MORE_ITEMS:
                    break
            values.append(name)
            i += 1
        for value in values:
            win32api.RegDeleteValue(key, value)

    orig_key = key
    try:
        key = win32api.RegOpenKeyEx(orig_key, subkey, 0, win32con.KEY_ALL_ACCESS)
    except pywintypes.error, e:
        if e[0] != winerror.ERROR_FILE_NOT_FOUND:
            raise
    else:
        delete_contents(key)
        win32api.RegCloseKey(key)
    try:
        win32api.RegDeleteKey(orig_key, subkey)
    except pywintypes.error, e:
        if e[0] == winerror.ERROR_FILE_NOT_FOUND:
            pass



# -*- coding: utf-8 -*-
"""add batchrow.status_text

Revision ID: 2dd756a44a16
Revises: 218562884128
Create Date: 2015-02-14 04:02:22.262221

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '2dd756a44a16'
down_revision = u'218562884128'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():
    op.add_column('vendor_catalog_row', sa.Column('status_text', sa.String(length=255), nullable=True))
    op.add_column('vendor_invoice_row', sa.Column('status_text', sa.String(length=255), nullable=True))


def downgrade():
    op.drop_column('vendor_invoice_row', 'status_text')
    op.drop_column('vendor_catalog_row', 'status_text')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Fabric Library for MySQL
"""

from __future__ import unicode_literals, absolute_import

from fabric.api import sudo, hide, get, put, abort, local, run
from fabric.contrib.files import sed

from rattail.fablib import apt, make_deploy


deploy = make_deploy(__file__)


def install(password):
    """
    Install the MySQL database service
    """
    deploy('mysql/debconf.mako', 'debconf', context={'password': password})
    sudo('debconf-set-selections debconf')
    sudo('rm debconf')
    apt.install('mysql-server')
    deploy('mysql/my.cnf.mako', '/root/.my.cnf', mode='0600', context={'password': password})


def is_mariadb():
    """
    Returns boolean indicating if MySQL server is actually MariaDB.
    """
    if "MariaDB" in run('mysql --version'):
        return True
    return False


def restart():
    """
    Restart the MySQL database service
    """
    sudo('service mysql restart')


def set_bind_address(address):
    """
    Configure the 'bind-address' setting with the given value.
    """
    sed('/etc/mysql/my.cnf', '^bind-address.*', 'bind-address = {}'.format(address), use_sudo=True)


def user_exists(name, host='localhost'):
    """
    Determine if a given MySQL user exists.
    """
    user = sql("SELECT User FROM user WHERE User = '{0}' and Host = '{1}'".format(name, host), database='mysql')
    return user == name


def create_user(name, host='localhost', password=None, checkfirst=True):
    """
    Create a MySQL user account.
    """
    if not checkfirst or not user_exists(name, host):
        sql("CREATE USER '{0}'@'{1}';".format(name, host))
    if password:
        with hide('running'):
            sql("SET PASSWORD FOR '{0}'@'{1}' = PASSWORD('{2}');".format(
                name, host, password))


def drop_user(name, host='localhost'):
    """
    Drop a MySQL user account.
    """
    sql("drop user '{0}'@'{1}'".format(name, host))
    

def db_exists(name):
    """
    Determine if a given MySQL database exists.
    """
    db = sql("SELECT SCHEMA_NAME FROM SCHEMATA WHERE SCHEMA_NAME = '{0}'".format(name), database='information_schema')
    return db == name


def table_exists(name, database):
    """
    Determine if a given table exists within the given MySQL database.
    """
    table = sql("SELECT TABLE_NAME FROM TABLES WHERE TABLE_SCHEMA = '{0}' AND TABLE_NAME = '{1}'".format(database, name), database='information_schema')
    return table == name


def create_db(name, checkfirst=True, user=None):
    """
    Create a MySQL database.
    """
    if not checkfirst or not db_exists(name):
        sudo('mysqladmin create {0}'.format(name))
        if user:
            grant_access(name, user)


def drop_db(name, checkfirst=True):
    """
    Drop a MySQL database.
    """
    if not checkfirst or db_exists(name):
        sudo('mysqladmin drop --force {0}'.format(name))


def grant_access(dbname, username):
    """
    Grant full access to the given database for the given user.  Note that the
    username should be given in MySQL's native format, e.g. 'myuser@localhost'.
    """
    sql('grant all on `{0}`.* to {1}'.format(dbname, username))


def sql(sql, database=''):
    """
    Execute some SQL.
    """
    # some crazy quoting required here, see also
    # http://stackoverflow.com/a/1250279
    sql = sql.replace("'", "'\"'\"'")
    return sudo("mysql --execute='{}' --batch --skip-column-names {}".format(sql, database))


def download_db(name, destination=None):
    """
    Download a database from the "current" server.
    """
    if destination is None:
        destination = './{0}.sql.gz'.format(name)
    sudo('mysqldump --result-file={0}.sql {0}'.format(name))
    sudo('gzip --force {0}.sql'.format(name))
    get('{0}.sql.gz'.format(name), destination)
    sudo('rm {0}.sql.gz'.format(name))


def restore_db(name, source=None, user=None):
    """
    Upload and restore a database to the current server
    """
    if not source:
        source = '{}.sql.gz'.format(name)
    put(source, '{}.sql.gz'.format(name))
    sudo('gunzip --force {}.sql.gz'.format(name))
    drop_db(name)
    create_db(name, user=user)
    sudo('mysql --execute="source {0}.sql" {0}'.format(name))
    sudo('rm {}.sql'.format(name))


def clone_db(name, download, force=False):
    """
    Clone a MySQL database from a (presumably live) server

    :param name: Name of the database.

    :param force: Whether the target database should be forcibly dropped, if it
       exists already.
    """
    if db_exists(name):
       if force:
           drop_db(name, checkfirst=False)
       else:
           abort("Database '{}' already exists! (pass force=true to override)".format(name))

    create_db(name, checkfirst=False)

    # obtain database dump from live server
    download()

    # upload database dump to target server
    put('{}.sql.gz'.format(name))
    local('rm {}.sql.gz'.format(name))

    # restore database on target server
    run('gunzip --force {}.sql.gz'.format(name))
    sudo('mysql {0} < {0}.sql'.format(name))
    run('rm {}.sql'.format(name))



# -*- coding: utf-8; -*-
# -*- coding: utf-8; -*-
"""
Views for ${model_title} batches
"""

from __future__ import unicode_literals, absolute_import

import six

from rattail.db import model

from tailbone.views.batch import FileBatchMasterView

from .handler import ${model_name}Handler


class ${model_name}View(FileBatchMasterView):
    """
    Master view for ${model_title} batches.
    """
    model_class = model.${model_name}
    batch_handler_class = ${model_name}Handler
    model_row_class = model.${model_name}Row
    model_title = "${model_title} Batch"
    model_title_plural = "${model_title} Batches"
    route_prefix = 'batch.${model_name.lower()}s'
    url_prefix = '/batch/${table_name.replace('_', '-')}'

    def get_instance_title(self, batch):
        return six.text_type(batch.vendor)

    def configure_grid(self, g):
        g.joiners['vendor'] = lambda q: q.join(model.Vendor)
        g.filters['vendor'] = g.make_filter('vendor', model.Vendor.name,
                                            default_active=True, default_verb='contains')
        g.sorters['vendor'] = g.make_sorter(model.Vendor.name)
        g.configure(
            include=[
                g.created,
                g.created_by,
                g.vendor,
                g.effective,
                g.filename,
                g.executed,
            ],
            readonly=True)

    def configure_fieldset(self, fs):
        fs.vendor.set(renderer=forms.renderers.VendorFieldRenderer)
        fs.configure(
            include=[
                fs.vendor.readonly(),
                fs.filename,
                fs.effective.readonly(),
                fs.created,
                fs.created_by,
                fs.executed,
                fs.executed_by,
            ])

    def configure_row_grid(self, g):
        g.configure(
            include=[
                g.sequence,
                g.upc.label("UPC"),
                g.description,
                g.status_code,
            ],
            readonly=True)

    def row_grid_row_attrs(self, row, i):
        attrs = {}
        if row.status_code == row.STATUS_SOME_CONCERN:
            attrs['class_'] = 'notice'
        if row.status_code == row.STATUS_UTTER_CHAOS:
            attrs['class_'] = 'warning'
        return attrs


def includeme(config):

    # fix permission group title
    config.add_tailbone_permission_group('${model_name.lower()}s', "${model_title}s")

    ${model_name}View.defaults(config)



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Common email config objects
"""

from __future__ import unicode_literals, absolute_import

import sys
import socket
from traceback import format_exception

from rattail.mail import Email


class datasync_error_watcher_get_changes(Email):
    """
    When any datasync watcher thread encounters an error trying to get changes,
    this email is sent out.
    """
    default_subject = "Watcher failed to get changes"

    def sample_data(self, request):
        from rattail.datasync import DataSyncWatcher
        try:
            raise RuntimeError("Fake error for preview")
        except:
            exc_type, exc, traceback = sys.exc_info()
        watcher = DataSyncWatcher(self.config, 'test')
        watcher.consumes_self = True
        return {
            'watcher': watcher,
            'error': exc,
            'traceback': ''.join(format_exception(exc_type, exc, traceback)).strip(),
            'datasync_url': '/datasyncchanges',
            'attempts': 2,
        }


class filemon_action_error(Email):
    """
    When any filemon thread encounters an error (and the retry attempts have
    been exhausted) then it will send out this email.
    """
    default_subject = "Error invoking action(s)"

    def sample_data(self, request):
        from rattail.filemon import Action
        action = Action(self.config)
        action.spec = 'rattail.filemon.actions:Action'
        action.retry_delay = 10
        try:
            raise RuntimeError("Fake error for preview")
        except:
            exc_type, exc, traceback = sys.exc_info()
        return {
            'hostname': socket.gethostname(),
            'path': '/tmp/foo.csv',
            'action': action,
            'attempts': 3,
            'error': exc,
            'traceback': ''.join(format_exception(exc_type, exc, traceback)).strip(),
        }


class rattail_import_versions_updates(Email):
    """
    Sent when a "version catch-up" import is performed, which involves changes.
    """
    fallback_key = 'rattail_import_updates'
    default_subject = "Changes for Rattail -> Rattail (Versions) import"


class user_feedback(Email):
    """
    Sent when a user submits a Feedback form from the web UI.
    """
    default_subject = "User Feedback"

    def sample_data(self, request):
        return {
            'user': None,
            'user_name': "Fred Flintstone",
            'referrer': request.route_url('home'),
            'message': "Hey there,\n\njust wondered what the heck was going on with this site?  It's crap.\n\nFred",
        }



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Rattail -> Rattail bulk data import
"""

from __future__ import unicode_literals, absolute_import

from rattail import importing
from rattail.util import OrderedDict
from rattail.importing.rattail import FromRattailToRattail, FromRattail


class BulkFromRattailToRattail(FromRattailToRattail, importing.BulkImportHandler):
    """
    Handler for Rattail -> Rattail bulk data import.
    """

    def get_importers(self):
        importers = OrderedDict()
        importers['Person'] = PersonImporter
        importers['PersonEmailAddress'] = PersonEmailAddressImporter
        importers['PersonPhoneNumber'] = PersonPhoneNumberImporter
        importers['PersonMailingAddress'] = PersonMailingAddressImporter
        importers['User'] = UserImporter
        importers['Message'] = MessageImporter
        importers['MessageRecipient'] = MessageRecipientImporter
        importers['Store'] = StoreImporter
        importers['StorePhoneNumber'] = StorePhoneNumberImporter
        importers['Employee'] = EmployeeImporter
        importers['EmployeeStore'] = EmployeeStoreImporter
        importers['EmployeeEmailAddress'] = EmployeeEmailAddressImporter
        importers['EmployeePhoneNumber'] = EmployeePhoneNumberImporter
        importers['ScheduledShift'] = ScheduledShiftImporter
        importers['WorkedShift'] = WorkedShiftImporter
        importers['Customer'] = CustomerImporter
        importers['CustomerGroup'] = CustomerGroupImporter
        importers['CustomerGroupAssignment'] = CustomerGroupAssignmentImporter
        importers['CustomerPerson'] = CustomerPersonImporter
        importers['CustomerEmailAddress'] = CustomerEmailAddressImporter
        importers['CustomerPhoneNumber'] = CustomerPhoneNumberImporter
        importers['Vendor'] = VendorImporter
        importers['VendorEmailAddress'] = VendorEmailAddressImporter
        importers['VendorPhoneNumber'] = VendorPhoneNumberImporter
        importers['VendorContact'] = VendorContactImporter
        importers['Department'] = DepartmentImporter
        importers['EmployeeDepartment'] = EmployeeDepartmentImporter
        importers['Subdepartment'] = SubdepartmentImporter
        importers['Category'] = CategoryImporter
        importers['Family'] = FamilyImporter
        importers['ReportCode'] = ReportCodeImporter
        importers['DepositLink'] = DepositLinkImporter
        importers['Tax'] = TaxImporter
        importers['Brand'] = BrandImporter
        importers['Product'] = ProductImporter
        importers['ProductCode'] = ProductCodeImporter
        importers['ProductCost'] = ProductCostImporter
        importers['ProductPrice'] = ProductPriceImporter
        return importers


class BulkFromRattail(FromRattail, importing.BulkToPostgreSQL):
    """
    Base class for bulk Rattail -> Rattail importers.
    """


class PersonImporter(BulkFromRattail, importing.model.PersonImporter):
    pass

class PersonEmailAddressImporter(BulkFromRattail, importing.model.PersonEmailAddressImporter):
    pass

class PersonPhoneNumberImporter(BulkFromRattail, importing.model.PersonPhoneNumberImporter):
    pass

class PersonMailingAddressImporter(BulkFromRattail, importing.model.PersonMailingAddressImporter):
    pass

class UserImporter(BulkFromRattail, importing.model.UserImporter):
    pass

class MessageImporter(BulkFromRattail, importing.model.MessageImporter):
    pass

class MessageRecipientImporter(BulkFromRattail, importing.model.MessageRecipientImporter):
    pass

class StoreImporter(BulkFromRattail, importing.model.StoreImporter):
    pass

class StorePhoneNumberImporter(BulkFromRattail, importing.model.StorePhoneNumberImporter):
    pass

class EmployeeImporter(BulkFromRattail, importing.model.EmployeeImporter):
    pass

class EmployeeStoreImporter(BulkFromRattail, importing.model.EmployeeStoreImporter):
    pass

class EmployeeDepartmentImporter(BulkFromRattail, importing.model.EmployeeDepartmentImporter):
    pass

class EmployeeEmailAddressImporter(BulkFromRattail, importing.model.EmployeeEmailAddressImporter):
    pass

class EmployeePhoneNumberImporter(BulkFromRattail, importing.model.EmployeePhoneNumberImporter):
    pass

class ScheduledShiftImporter(BulkFromRattail, importing.model.ScheduledShiftImporter):
    pass

class WorkedShiftImporter(BulkFromRattail, importing.model.WorkedShiftImporter):
    pass

class CustomerImporter(BulkFromRattail, importing.model.CustomerImporter):
    pass

class CustomerGroupImporter(BulkFromRattail, importing.model.CustomerGroupImporter):
    pass

class CustomerGroupAssignmentImporter(BulkFromRattail, importing.model.CustomerGroupAssignmentImporter):
    pass

class CustomerPersonImporter(BulkFromRattail, importing.model.CustomerPersonImporter):
    pass

class CustomerEmailAddressImporter(BulkFromRattail, importing.model.CustomerEmailAddressImporter):
    pass

class CustomerPhoneNumberImporter(BulkFromRattail, importing.model.CustomerPhoneNumberImporter):
    pass

class VendorImporter(BulkFromRattail, importing.model.VendorImporter):
    pass

class VendorEmailAddressImporter(BulkFromRattail, importing.model.VendorEmailAddressImporter):
    pass

class VendorPhoneNumberImporter(BulkFromRattail, importing.model.VendorPhoneNumberImporter):
    pass

class VendorContactImporter(BulkFromRattail, importing.model.VendorContactImporter):
    pass

class DepartmentImporter(BulkFromRattail, importing.model.DepartmentImporter):
    pass

class SubdepartmentImporter(BulkFromRattail, importing.model.SubdepartmentImporter):
    pass

class CategoryImporter(BulkFromRattail, importing.model.CategoryImporter):
    pass

class FamilyImporter(BulkFromRattail, importing.model.FamilyImporter):
    pass

class ReportCodeImporter(BulkFromRattail, importing.model.ReportCodeImporter):
    pass

class DepositLinkImporter(BulkFromRattail, importing.model.DepositLinkImporter):
    pass

class TaxImporter(BulkFromRattail, importing.model.TaxImporter):
    pass

class BrandImporter(BulkFromRattail, importing.model.BrandImporter):
    pass


class ProductImporter(BulkFromRattail, importing.model.ProductImporter):
    """
    Product data requires some extra handling currently.  The bulk importer
    does not support the regular/current price foreign key fields, so those
    must be populated in some other way after the initial bulk import.
    """

    @property
    def simple_fields(self):
        fields = super(ProductImporter, self).simple_fields
        fields.remove('regular_price_uuid')
        fields.remove('current_price_uuid')
        return fields


class ProductCodeImporter(BulkFromRattail, importing.model.ProductCodeImporter):
    pass

class ProductCostImporter(BulkFromRattail, importing.model.ProductCostImporter):
    pass

class ProductPriceImporter(BulkFromRattail, importing.model.ProductPriceImporter):
    pass



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data Models for Users & Permissions
"""

from __future__ import unicode_literals, absolute_import

import datetime

import six
import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.ext.associationproxy import association_proxy

from rattail.db.model import Base, uuid_column, getset_factory, Person
 

@six.python_2_unicode_compatible
class Role(Base):
    """
    Represents a role within the system; used to manage permissions.
    """
    __tablename__ = 'role'
    __table_args__ = (
        sa.UniqueConstraint('name', name='role_uq_name'),
        )
    __versioned__ = {}

    uuid = uuid_column()
    name = sa.Column(sa.String(length=25), nullable=False)

    session_timeout = sa.Column(sa.Integer(), nullable=True, doc="""
    Optional session timeout value for the role, in seconds.  If this is set to
    zero, the role's users will have no session timeout.  A value of ``None``
    means the role has no say in the timeout.
    """)

    def __str__(self):
        return self.name or ''


@six.python_2_unicode_compatible
class Permission(Base):
    """
    Represents permission a role has to do a particular thing.
    """
    __tablename__ = 'permission'
    __table_args__ = (
        sa.ForeignKeyConstraint(['role_uuid'], ['role.uuid'], name='permission_fk_role'),
        )
    # __versioned__ = {}

    role_uuid = sa.Column(sa.String(length=32), primary_key=True)
    permission = sa.Column(sa.String(length=50), primary_key=True)

    def __str__(self):
        return self.permission or ''


Role._permissions = orm.relationship(
    Permission, backref='role',
    cascade='save-update, merge, delete, delete-orphan')

Role.permissions = association_proxy(
    '_permissions', 'permission',
    creator=lambda p: Permission(permission=p),
    getset_factory=getset_factory)


@six.python_2_unicode_compatible
class User(Base):
    """
    Represents a user of the system.

    This may or may not correspond to a real person, i.e. some users may exist
    solely for automated tasks.
    """
    __tablename__ = 'user'
    __table_args__ = (
        sa.ForeignKeyConstraint(['person_uuid'], ['person.uuid'], name='user_fk_person'),
        sa.UniqueConstraint('username', name='user_uq_username'),
        )
    __versioned__ = {'exclude': ['password', 'salt']}

    uuid = uuid_column()
    username = sa.Column(sa.String(length=25), nullable=False)
    password = sa.Column(sa.String(length=60))
    salt = sa.Column(sa.String(length=29))

    person_uuid = sa.Column(sa.String(length=32))
    person = orm.relationship(
        Person,
        uselist=False,
        backref=orm.backref(
            'user',
            uselist=False))

    active = sa.Column(sa.Boolean(), nullable=False, default=True, doc="""
    Whether the user is active, e.g. allowed to log in via the UI.
    """)

    active_sticky = sa.Column(sa.Boolean(), nullable=True, doc="""
    Optional flag, motivation behind which is as follows: If you import user
    accounts from another system, esp. on a regular basis, you might be keeping
    the :attr:`active` flag in sync along with that.  But in some cases you
    might want to *not* keep the active flag in sync, for certain accounts.
    Hence this "active sticky" flag, which may be used to mark certain accounts
    as off-limits from the general active flag sync.
    """)

    def __str__(self):
        if self.person and str(self.person):
            return str(self.person)
        return self.username or ''

    @property
    def display_name(self):
        """
        Display name for the user.
        
        Returns :attr:`Person.display_name` if available; otherwise returns
        :attr:`username`.
        """
        if self.person and self.person.display_name:
            return self.person.display_name
        return self.username

    @property
    def employee(self):
        """
        Reference to the :class:`Employee` associated with the user, if any.
        """
        if self.person:
            return self.person.employee

    def get_short_name(self):
        """
        Returns "short name" for the user.  This is for convenience of mobile
        view, at least...
        """
        # TODO: this should reference employee.short_name
        employee = self.employee
        if employee and employee.display_name:
            return employee.display_name

        person = self.person
        if person:
            if person.first_name and person.last_name:
                return "{} {}.".format(person.first_name, person.last_name[0])
            if person.first_name:
                return person.first_name

        return self.username

    def get_email_address(self):
        """
        Returns the primary email address for the user (as unicode string), or
        ``None``.  Note that currently there is no direct association between a
        User and an EmailAddress, so the Person and Customer relationships are
        navigated in an attempt to locate an address.
        """
        if self.person:
            if self.person.email:
                return self.person.email.address
            for customer in self.person.customers:
                if customer.email:
                    return customer.email.address

    @property
    def email_address(self):
        """
        Convenience attribute which invokes :meth:`get_email_address()`.

        .. note::
           The implementation of this may change some day, e.g. if the User is
           given an association to EmailAddress in the data model.
        """
        return self.get_email_address()

    def is_admin(self):
        """
        Convenience method to determine if current user is a member of the
        Administrator role.
        """
        from rattail.db.auth import administrator_role

        session = orm.object_session(self)
        return administrator_role(session) in self.roles

    def record_event(self, type_code, **kwargs):
        kwargs['type_code'] = type_code
        self.events.append(UserEvent(**kwargs))


Person.make_proxy(User, 'person', 'first_name')
Person.make_proxy(User, 'person', 'last_name')
Person.make_proxy(User, 'person', 'display_name')


class UserRole(Base):
    """
    Represents the association between a :class:`User` and a :class:`Role`.
    """
    __tablename__ = 'user_x_role'
    __table_args__ = (
        sa.ForeignKeyConstraint(['user_uuid'], ['user.uuid'], name='user_x_role_fk_user'),
        sa.ForeignKeyConstraint(['role_uuid'], ['role.uuid'], name='user_x_role_fk_role'),
        )
    __versioned__ = {}

    uuid = uuid_column()
    user_uuid = sa.Column(sa.String(length=32), nullable=False)
    role_uuid = sa.Column(sa.String(length=32), nullable=False)


Role._users = orm.relationship(
    UserRole, backref='role',
    cascade='all, delete-orphan')

Role.users = association_proxy(
    '_users', 'user',
    creator=lambda u: UserRole(user=u),
    getset_factory=getset_factory)

User._roles = orm.relationship(
    UserRole, backref='user',
    cascade='all, delete-orphan')

User.roles = association_proxy(
    '_roles', 'role',
    creator=lambda r: UserRole(role=r),
    getset_factory=getset_factory)


class UserEvent(Base):
    """
    Represents an event associated with a user.
    """
    __tablename__ = 'user_event'
    __table_args__ = (
        sa.ForeignKeyConstraint(['user_uuid'], ['user.uuid'], name='user_x_role_fk_user'),
    )

    uuid = uuid_column()

    user_uuid = sa.Column(sa.String(length=32), nullable=False)
    user = orm.relationship(
        User,
        doc="""
        Reference to the user whose event this is.
        """,
        backref=orm.backref(
            'events',
            doc="""
            Sequence of events for the user.
            """))

    type_code = sa.Column(sa.Integer(), nullable=False, doc="""
    Type code for the event.
    """)

    occurred = sa.Column(sa.DateTime(), nullable=True, default=datetime.datetime.utcnow, doc="""
    Timestamp at which the event occurred.
    """)



# -*- coding: utf-8 -*-
"""add purchase damaged, expired counts

Revision ID: 123af5f6a0bc
Revises: 4bd528bca236
Create Date: 2016-12-09 14:44:04.728249

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '123af5f6a0bc'
down_revision = u'4bd528bca236'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


batch_id_seq = sa.Sequence('batch_id_seq')


def upgrade():

    # purchase_batch_row
    op.add_column('purchase_batch_row', sa.Column('cases_damaged', sa.Numeric(precision=10, scale=4), nullable=True))
    op.add_column('purchase_batch_row', sa.Column('cases_expired', sa.Numeric(precision=10, scale=4), nullable=True))
    op.add_column('purchase_batch_row', sa.Column('units_damaged', sa.Numeric(precision=10, scale=4), nullable=True))
    op.add_column('purchase_batch_row', sa.Column('units_expired', sa.Numeric(precision=10, scale=4), nullable=True))

    # purchase_item
    op.add_column('purchase_item', sa.Column('cases_damaged', sa.Numeric(precision=10, scale=4), nullable=True))
    op.add_column('purchase_item', sa.Column('cases_expired', sa.Numeric(precision=10, scale=4), nullable=True))
    op.add_column('purchase_item', sa.Column('units_damaged', sa.Numeric(precision=10, scale=4), nullable=True))
    op.add_column('purchase_item', sa.Column('units_expired', sa.Numeric(precision=10, scale=4), nullable=True))


def downgrade():

    # purchase_item
    op.drop_column('purchase_item', 'units_expired')
    op.drop_column('purchase_item', 'units_damaged')
    op.drop_column('purchase_item', 'cases_expired')
    op.drop_column('purchase_item', 'cases_damaged')

    # purchase_batch_row
    op.drop_column('purchase_batch_row', 'units_expired')
    op.drop_column('purchase_batch_row', 'units_damaged')
    op.drop_column('purchase_batch_row', 'cases_expired')
    op.drop_column('purchase_batch_row', 'cases_damaged')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Import Handlers
"""

from __future__ import unicode_literals, absolute_import

import sys
import logging

import six
import humanize

from rattail.time import make_utc
from rattail.util import OrderedDict
from rattail.mail import send_email


log = logging.getLogger(__name__)


class ImportHandler(object):
    """
    Base class for all import handlers.
    """
    host_title = None
    local_title = None
    progress = None
    dry_run = False
    commit_host_partial = False
    diff_max_display = 15

    def __init__(self, config=None, **kwargs):
        self.config = config
        self.enum = config.get_enum() if config else None
        self.importers = self.get_importers()
        for key, value in kwargs.iteritems():
            setattr(self, key, value)

    def get_importers(self):
        """
        Returns a dict of all available importers, where the keys are model
        names and the values are importer factories.  All subclasses will want
        to override this.  Note that if you return an
        :class:`python:collections.OrderedDict` instance, you can affect the
        ordering of keys in the command line help system, etc.
        """
        return {}

    def get_importer_keys(self):
        """
        Returns the list of keys corresponding to the available importers.
        """
        return list(self.importers.iterkeys())

    def get_default_keys(self):
        """
        Returns the list of keys corresponding to the "default" importers.
        Override this if you wish certain importers to be excluded by default,
        e.g. when first testing them out etc.
        """
        return self.get_importer_keys()

    def get_importer(self, key, **kwargs):
        """
        Returns an importer instance corresponding to the given key.
        """
        if key in self.importers:
            kwargs.setdefault('handler', self)
            kwargs.setdefault('config', self.config)
            kwargs.setdefault('host_system_title', self.host_title)
            if hasattr(self, 'batch_size'):
                kwargs.setdefault('batch_size', self.batch_size)
            kwargs = self.get_importer_kwargs(key, **kwargs)
            return self.importers[key](**kwargs)

    def get_importer_kwargs(self, key, **kwargs):
        """
        Return a dict of kwargs to be used when construcing an importer with
        the given key.
        """
        return kwargs

    def import_data(self, *keys, **kwargs):
        """
        Import all data for the given importer/model keys.
        """
        self.import_began = make_utc(tzinfo=True)
        if 'dry_run' in kwargs:
            self.dry_run = kwargs['dry_run']
        self.progress = kwargs.pop('progress', getattr(self, 'progress', None))
        self.warnings = kwargs.get('warnings', False)
        kwargs.update({'dry_run': self.dry_run,
                       'progress': self.progress})

        self.setup()
        self.begin_transaction()
        changes = OrderedDict()

        try:
            for key in keys:
                importer = self.get_importer(key, **kwargs)
                if importer:
                    created, updated, deleted = importer.import_data()
                    changed = bool(created or updated or deleted)
                    logger = log.warning if changed and self.warnings else log.info
                    logger("{} -> {}: added {:,d}; updated {:,d}; deleted {:,d} {} records".format(
                        self.host_title, self.local_title, len(created), len(updated), len(deleted), key))
                    if changed:
                        changes[key] = created, updated, deleted
                else:
                    log.warning("skipping unknown importer: {}".format(key))
        except:
            if self.commit_host_partial and not self.dry_run:
                log.warning("{host} -> {local}: committing partial transaction on host {host} (despite error)".format(
                    host=self.host_title, local=self.local_title))
                self.commit_host_transaction()
            raise
        else:
            if changes:
                self.process_changes(changes)
            if self.dry_run:
                self.rollback_transaction()
            else:
                self.commit_transaction()

        self.teardown()
        return changes

    def setup(self):
        """
        Perform any additional setup if/as necessary, prior to running the
        import task(s).
        """

    def teardown(self):
        """
        Perform any cleanup necessary, after running the import task(s).
        """

    def begin_transaction(self):
        self.begin_host_transaction()
        self.begin_local_transaction()

    def begin_host_transaction(self):
        pass

    def begin_local_transaction(self):
        pass

    def rollback_transaction(self):
        self.rollback_host_transaction()
        self.rollback_local_transaction()

    def rollback_host_transaction(self):
        pass

    def rollback_local_transaction(self):
        pass

    def commit_transaction(self):
        self.commit_host_transaction()
        self.commit_local_transaction()

    def commit_host_transaction(self):
        pass

    def commit_local_transaction(self):
        pass

    def process_changes(self, changes):
        """
        This method is called any time changes occur, regardless of whether the
        import is running in "warnings" mode.  Default implementation does
        nothing; override as needed.
        """
        # TODO: This whole thing needs a re-write...but for now, waiting until
        # the old importer has really gone away, so we can share its email
        # template instead of bothering with something more complicated.

        if not self.warnings:
            return

        now = make_utc(tzinfo=True)
        data = {
            'local_title': self.local_title,
            'host_title': self.host_title,
            'argv': sys.argv,
            'runtime': humanize.naturaldelta(now - self.import_began),
            'changes': changes,
            'dry_run': self.dry_run,
            'render_record': RecordRenderer(self.config),
            'max_display': self.diff_max_display,
        }

        command = getattr(self, 'command', None)
        if command:
            data['command'] = '{} {}'.format(command.parent.name, command.name)
        else:
            data['command'] = None

        if command:
            key = '{}_{}_updates'.format(command.parent.name, command.name)
            key = key.replace('-', '_')
        else:
            key = 'rattail_import_updates'

        send_email(self.config, key, fallback_key='rattail_import_updates', data=data)
        log.info("{} -> {}: warning email was sent".format(self.host_title, self.local_title))


class BulkImportHandler(ImportHandler):
    """
    Base class for bulk import handlers.
    """

    def import_data(self, *keys, **kwargs):
        """
        Import all data for the given importer/model keys.
        """
        # TODO: still need to refactor much of this so can share with parent class
        self.import_began = make_utc(tzinfo=True)
        if 'dry_run' in kwargs:
            self.dry_run = kwargs['dry_run']
        self.progress = kwargs.pop('progress', getattr(self, 'progress', None))
        self.warnings = kwargs.pop('warnings', False)
        kwargs.update({'dry_run': self.dry_run,
                       'progress': self.progress})
        self.setup()
        self.begin_transaction()
        changes = OrderedDict()

        try:
            for key in keys:
                importer = self.get_importer(key, **kwargs)
                if not importer:
                    log.warning("skipping unknown importer: {}".format(key))
                    continue

                created = importer.import_data()
                log.info("{} -> {}: added {}, updated 0, deleted 0 {} records".format(
                    self.host_title, self.local_title, created, key))
                if created:
                    changes[key] = created
        except:
            if self.commit_host_partial and not self.dry_run:
                log.warning("{host} -> {local}: committing partial transaction on host {host} (despite error)".format(
                    host=self.host_title, local=self.local_title))
                self.commit_host_transaction()
            raise
        else:
            if self.dry_run:
                self.rollback_transaction()
            else:
                self.commit_transaction()

        self.teardown()
        return changes


class FromSQLAlchemyHandler(ImportHandler):
    """
    Handler for imports for which the host data source is represented by a
    SQLAlchemy engine and ORM.
    """
    host_session = None

    def make_host_session(self):
        """
        Subclasses must override this to define the host database connection.
        """
        raise NotImplementedError

    def get_importer_kwargs(self, key, **kwargs):
        kwargs = super(FromSQLAlchemyHandler, self).get_importer_kwargs(key, **kwargs)
        kwargs.setdefault('host_session', self.host_session)
        return kwargs

    def begin_host_transaction(self):
        self.host_session = self.make_host_session()

    def rollback_host_transaction(self):
        self.host_session.rollback()
        self.host_session.close()
        self.host_session = None

    def commit_host_transaction(self):
        self.host_session.commit()
        self.host_session.close()
        self.host_session = None


class ToSQLAlchemyHandler(ImportHandler):
    """
    Handler for imports which target a SQLAlchemy ORM on the local side.
    """
    session = None

    def make_session(self):
        """
        Subclasses must override this to define the local database connection.
        """
        raise NotImplementedError

    def get_importer_kwargs(self, key, **kwargs):
        kwargs = super(ToSQLAlchemyHandler, self).get_importer_kwargs(key, **kwargs)
        kwargs.setdefault('session', self.session)
        return kwargs

    def begin_local_transaction(self):
        self.session = self.make_session()

    def rollback_local_transaction(self):
        self.session.rollback()
        self.session.close()
        self.session = None

    def commit_local_transaction(self):
        self.session.commit()
        self.session.close()
        self.session = None


# TODO: this is still pretty hacky, but needed to get it in here for now
class RecordRenderer(object):
    """
    Record renderer for email notifications sent from data import jobs.
    """

    def __init__(self, config):
        self.config = config

    def __call__(self, record):
        return self.render(record)

    def render(self, record):
        """
        Render the given record.
        """
        key = record.__class__.__name__.lower()
        renderer = getattr(self, 'render_{}'.format(key), None)
        if renderer:
            return renderer(record)

        label = self.get_label(record)
        url = self.get_url(record)
        if url:
            return '<a href="{}">{}</a>'.format(url, label)
        return label

    def get_label(self, record):
        key = record.__class__.__name__.lower()
        label = getattr(self, 'label_{}'.format(key), self.label)
        return label(record)

    def label(self, record):
        return six.text_type(record)

    def get_url(self, record):
        """
        Fetch / generate a URL for the given data record.  You should *not*
        override this method, but do :meth:`url()` instead.
        """
        key = record.__class__.__name__.lower()
        url = getattr(self, 'url_{}'.format(key), self.url)
        return url(record)

    def url(self, record):
        """
        Fetch / generate a URL for the given data record.
        """
        if hasattr(record, 'uuid'):
            url = self.config.get('tailbone', 'url')
            if url:
                url = url.rstrip('/')
                name = '{}s'.format(record.__class__.__name__.lower())
                if name == 'persons': # FIXME, obviously this is a hack
                    name = 'people'
                url = '{}/{}/{{uuid}}'.format(url, name)
                return url.format(uuid=record.uuid)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
appy.pod integration
"""

from __future__ import unicode_literals, absolute_import

import warnings


def render_document(config, template, context, output, **kwargs):
    """
    Render a document using appy.pod
    """
    from appy.pod.renderer import Renderer

    kwargs.setdefault('pythonWithUnoPath', config.get('rattail', 'uno_python'))
    renderer = Renderer(template, context, output, **kwargs)
    renderer.run()


def suppress_md5_warning():
    """
    Apply a filter to ignore some md5-related warning which comes up whenever
    we need to ``import appy``.
    """
    warnings.filterwarnings(
        'ignore',
        r"^the md5 module is deprecated; use hashlib instead$",
        DeprecationWarning, r'^appy\.shared\.dav$')



# -*- coding: utf-8; -*-
"""
TrainWreck data models
"""

from __future__ import unicode_literals, absolute_import

import six
import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.ext.declarative import declarative_base, declared_attr
from sqlalchemy.ext.orderinglist import ordering_list

from rattail import enum
from rattail.db.model import uuid_column
from rattail.db.model.core import ModelBase


Base = declarative_base(cls=ModelBase)


@six.python_2_unicode_compatible
class TransactionBase(object):
    """
    Represents a POS (or similar?) transaction.
    """
    __tablename__ = 'transaction'

    uuid = uuid_column()

    system = sa.Column(sa.String(length=50), nullable=True, doc="""
    Unique string representing the system from which transaction originated.
    """)

    terminal_id = sa.Column(sa.String(length=20), nullable=True, doc="""
    Terminal ID from which the transaction originated, if known.
    """)

    receipt_number = sa.Column(sa.String(length=20), nullable=True, doc="""
    Receipt number for the transaction, if known.
    """)

    start_time = sa.Column(sa.DateTime(), nullable=True, doc="""
    UTC timestamp when the transaction began, if known.
    """)

    end_time = sa.Column(sa.DateTime(), nullable=True, doc="""
    UTC timestamp when the transaction ended, if known.
    """)

    customer_id = sa.Column(sa.String(length=20), nullable=True, doc="""
    Customer ID string for the transaction, if known.  This should correspond
    to a proper customer account, as opposed to :attr:`shopper_id`.
    """)

    customer_name = sa.Column(sa.String(length=255), nullable=True, doc="""
    Customer name for the transaction, if known.  This should correspond to a
    proper customer account, as opposed to :attr:`shopper_name`.
    """)

    shopper_id = sa.Column(sa.String(length=20), nullable=True, doc="""
    Shopper ID string for the transaction, if known.  This should correspond
    to the actual shopper, as opposed to :attr:`customer_id`.
    """)

    shopper_name = sa.Column(sa.String(length=255), nullable=True, doc="""
    Shopper name for the transaction, if known.  This should correspond to the
    actual shopper, as opposed to :attr:`customer_name`.
    """)

    subtotal = sa.Column(sa.Numeric(precision=9, scale=2), nullable=True, doc="""
    Subtotal for the transaction, before discounts.
    """)

    discounted_subtotal = sa.Column(sa.Numeric(precision=9, scale=2), nullable=True, doc="""
    Subtotal for the transaction, after discounts.
    """)

    tax = sa.Column(sa.Numeric(precision=8, scale=2), nullable=True, doc="""
    Tax total (for all tax levels) for the transaction.
    """)

    tax1 = sa.Column(sa.Numeric(precision=8, scale=2), nullable=True, doc="""
    Total of "tax level 1" for the transaction.
    """)

    tax2 = sa.Column(sa.Numeric(precision=8, scale=2), nullable=True, doc="""
    Total of "tax level 2" for the transaction.
    """)

    tax3 = sa.Column(sa.Numeric(precision=8, scale=2), nullable=True, doc="""
    Total of "tax level 3" for the transaction.
    """)

    tax4 = sa.Column(sa.Numeric(precision=8, scale=2), nullable=True, doc="""
    Total of "tax level 4" for the transaction.
    """)

    tax5 = sa.Column(sa.Numeric(precision=8, scale=2), nullable=True, doc="""
    Total of "tax level 5" for the transaction.
    """)

    tax6 = sa.Column(sa.Numeric(precision=8, scale=2), nullable=True, doc="""
    Total of "tax level 6" for the transaction.
    """)

    tax7 = sa.Column(sa.Numeric(precision=8, scale=2), nullable=True, doc="""
    Total of "tax level 7" for the transaction.
    """)

    total = sa.Column(sa.Numeric(precision=9, scale=2), nullable=True, doc="""
    Total amount for the transaction, i.e. discounted subtotal plus tax.
    """)

    void = sa.Column(sa.Boolean(), nullable=False, default=False, doc="""
    Flag indicating if the transaction was voided.
    """)

    def __str__(self):
        system = enum.TRAINWRECK_SYSTEM.get(self.system, self.system or "Unknown")
        return "{} {}-{}".format(system, self.terminal_id or "?", self.receipt_number or "??")


@six.python_2_unicode_compatible
class TransactionItemBase(object):
    """
    Represents a line item within a transaction.
    """
    __tablename__ = 'transaction_item'

    @declared_attr
    def __table_args__(cls):
        return cls.__txnitem_table_args__()

    @classmethod
    def __txnitem_table_args__(cls):
        txn_table = cls.__txn_class__.__tablename__
        item_table = cls.__tablename__
        return (
            sa.ForeignKeyConstraint(['transaction_uuid'], ['{}.uuid'.format(txn_table)],
                                    name='{}_fk_transaction'.format(item_table)),
        )

    uuid = uuid_column()

    transaction_uuid = sa.Column(sa.String(length=32), nullable=False)

    @declared_attr
    def transaction(cls):
        txn_class = cls.__txn_class__
        item_class = cls
        txn_class.item_class = item_class

        # Must establish `Transaction.items` here instead of from within
        # Transaction itself, because the item class doesn't yet exist (really,
        # it can't even be "known") when Transaction is being created.
        txn_class.items = orm.relationship(
            item_class,
            order_by=lambda: item_class.sequence,
            collection_class=ordering_list('sequence', count_from=1),
            cascade='all, delete-orphan',
            doc="""
            Sequence of line items for the transaction.
            """,
            back_populates='transaction')

        # Now, here's the `TransactionItem.transaction` reference.
        return orm.relationship(txn_class, back_populates='items', doc="""
        Reference to the line item's parent transaction.
        """)

    sequence = sa.Column(sa.Integer(), nullable=False, doc="""
    Sequence number for this line item, within the parent transaction.
    """)

    item_type = sa.Column(sa.String(length=50), nullable=True, doc="""
    Type of line item this is, if relevant/known.  Value and interpretation
    thereof are not defined by the schema, but rather the (custom) app logic...
    """)

    item_id = sa.Column(sa.String(length=20), nullable=False, doc="""
    Item ID for the line item.  This is generally the product UPC, but may vary.
    """)

    department_number = sa.Column(sa.Integer(), nullable=True, doc="""
    Department number for the line item, if known.
    """)

    department_name = sa.Column(sa.String(length=30), nullable=True, doc="""
    Department name for the line item, if known.
    """)

    brand_name = sa.Column(sa.String(length=100), nullable=True, doc="""
    Brand name for the line item, if known.
    """)

    description = sa.Column(sa.String(length=255), nullable=False, default="", doc="""
    Description for the line item; defaults to empty string.
    """)

    unit_price = sa.Column(sa.Numeric(precision=8, scale=2), nullable=True, doc="""
    Effective unit price for the item, before discounts.
    """)

    unit_regular_price = sa.Column(sa.Numeric(precision=8, scale=2), nullable=True, doc="""
    Regular unit price for the item, e.g. as opposed to effective price.
    """)

    unit_discounted_price = sa.Column(sa.Numeric(precision=8, scale=2), nullable=True, doc="""
    Effective unit price for the item, after discounts.
    """)

    unit_quantity = sa.Column(sa.Numeric(precision=10, scale=4), nullable=True, doc="""
    Unity quantity for the line item, i.e. how much of product is being sold.
    """)

    subtotal = sa.Column(sa.Numeric(precision=9, scale=2), nullable=True, doc="""
    Subtotal for the line item, i.e. unit quantity times price, before discounts.
    """)

    discounted_subtotal = sa.Column(sa.Numeric(precision=9, scale=2), nullable=True, doc="""
    Subtotal for the line item, i.e. quantity times unit price.
    """)

    gross_sales = sa.Column(sa.Numeric(precision=9, scale=2), nullable=True, doc="""
    Gross sales amount for the line item.  Meaning of this may vary slightly?
    """)

    net_sales = sa.Column(sa.Numeric(precision=9, scale=2), nullable=True, doc="""
    Net sales amount for the line item.  Meaning of this may vary slightly?
    """)

    tax = sa.Column(sa.Numeric(precision=8, scale=2), nullable=True, doc="""
    Tax total (for all tax levels) for the line item.
    """)

    tax1 = sa.Column(sa.Numeric(precision=8, scale=2), nullable=True, doc="""
    Total of "tax level 1" for the line item.
    """)

    tax2 = sa.Column(sa.Numeric(precision=8, scale=2), nullable=True, doc="""
    Total of "tax level 2" for the line item.
    """)

    tax3 = sa.Column(sa.Numeric(precision=8, scale=2), nullable=True, doc="""
    Total of "tax level 3" for the line item.
    """)

    tax4 = sa.Column(sa.Numeric(precision=8, scale=2), nullable=True, doc="""
    Total of "tax level 4" for the line item.
    """)

    tax5 = sa.Column(sa.Numeric(precision=8, scale=2), nullable=True, doc="""
    Total of "tax level 5" for the line item.
    """)

    tax6 = sa.Column(sa.Numeric(precision=8, scale=2), nullable=True, doc="""
    Total of "tax level 6" for the line item.
    """)

    tax7 = sa.Column(sa.Numeric(precision=8, scale=2), nullable=True, doc="""
    Total of "tax level 7" for the line item.
    """)

    total = sa.Column(sa.Numeric(precision=9, scale=2), nullable=True, doc="""
    Total amount for the line item, i.e. discounted subtotal plus tax.
    """)

    void = sa.Column(sa.Boolean(), nullable=False, default=False, doc="""
    Flag indicating if the line item was voided.
    """)

    error_correct = sa.Column(sa.Boolean(), nullable=False, default=False, doc="""
    Flag indicating if the line item was error-corrected.
    """)

    def __str__(self):
        return self.description



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Importing Commands
"""

from __future__ import unicode_literals, absolute_import

import sys
import logging

import six

from rattail.commands.core import Subcommand, date_argument
from rattail.util import load_object


log = logging.getLogger(__name__)


class ImportSubcommand(Subcommand):
    """
    Base class for subcommands which use the (new) data importing system.
    """
    handler_spec = None

    # TODO: move this into Subcommand or something..
    parent_name = None
    def __init__(self, *args, **kwargs):
        if 'handler_spec' in kwargs:
            self.handler_spec = kwargs.pop('handler_spec')
        super(ImportSubcommand, self).__init__(*args, **kwargs)
        if self.parent:
            self.parent_name = self.parent.name

    def get_handler_factory(self, **kwargs):
        """
        Subclasses must override this, and return a callable that creates an
        import handler instance which the command should use.
        """
        if self.handler_spec:
            return load_object(self.handler_spec)
        raise NotImplementedError

    def get_handler(self, **kwargs):
        """
        Returns a handler instance to be used by the command.
        """
        factory = self.get_handler_factory(args=kwargs.get('args'))
        kwargs.setdefault('config', getattr(self, 'config', None))
        kwargs.setdefault('command', self)
        kwargs.setdefault('progress', self.progress)
        if hasattr(self, 'runas_username'):
            kwargs.setdefault('runas_username', self.runas_username)
            kwargs.setdefault('runas_user', self.get_runas_user())
        if 'args' in kwargs:
            args = kwargs['args']
            kwargs.setdefault('dry_run', args.dry_run)
            if hasattr(args, 'batch_size'):
                kwargs.setdefault('batch_size', args.batch_size)
            # kwargs.setdefault('max_create', args.max_create)
            # kwargs.setdefault('max_update', args.max_update)
            # kwargs.setdefault('max_delete', args.max_delete)
            # kwargs.setdefault('max_total', args.max_total)
        kwargs = self.get_handler_kwargs(**kwargs)
        return factory(**kwargs)

    def get_handler_kwargs(self, **kwargs):
        """
        Return a dict of kwargs to be passed to the handler factory.
        """
        return kwargs

    def add_parser_args(self, parser):

        # model names (aka importer keys)
        doc = ("Which data models to import.  If you specify any, then only "
               "data for those models will be imported.  If you do not specify "
               "any, then all *default* models will be imported.")
        try:
            handler = self.get_handler()
        except NotImplementedError:
            pass
        else:
            doc += "  Supported models are: ({})".format(', '.join(handler.get_importer_keys()))
        parser.add_argument('models', nargs='*', metavar='MODEL', help=doc)

        # start/end date
        parser.add_argument('--start-date', type=date_argument, metavar='DATE',
                            help="Optional (inclusive) starting point for date range, by which host "
                            "data should be filtered.  Only used by certain importers.")
        parser.add_argument('--end-date', type=date_argument, metavar='DATE',
                            help="Optional (inclusive) ending point for date range, by which host "
                            "data should be filtered.  Only used by certain importers.")

        # allow create?
        parser.add_argument('--create', action='store_true', default=True,
                            help="Allow new records to be created during the import.")
        parser.add_argument('--no-create', action='store_false', dest='create',
                            help="Do not allow new records to be created during the import.")
        parser.add_argument('--max-create', type=int, metavar='COUNT',
                            help="Maximum number of records which may be created, after which a "
                            "given import task should stop.  Note that this applies on a per-model "
                            "basis and not overall.")

        # allow update?
        parser.add_argument('--update', action='store_true', default=True,
                            help="Allow existing records to be updated during the import.")
        parser.add_argument('--no-update', action='store_false', dest='update',
                            help="Do not allow existing records to be updated during the import.")
        parser.add_argument('--max-update', type=int, metavar='COUNT',
                            help="Maximum number of records which may be updated, after which a "
                            "given import task should stop.  Note that this applies on a per-model "
                            "basis and not overall.")

        # allow delete?
        parser.add_argument('--delete', action='store_true', default=False,
                            help="Allow records to be deleted during the import.")
        parser.add_argument('--no-delete', action='store_false', dest='delete',
                            help="Do not allow records to be deleted during the import.")
        parser.add_argument('--max-delete', type=int, metavar='COUNT',
                            help="Maximum number of records which may be deleted, after which a "
                            "given import task should stop.  Note that this applies on a per-model "
                            "basis and not overall.")

        # max total changes, per model
        parser.add_argument('--max-total', type=int, metavar='COUNT',
                            help="Maximum number of *any* record changes which may occur, after which "
                            "a given import task should stop.  Note that this applies on a per-model "
                            "basis and not overall.")

        # batch size
        parser.add_argument('--batch', type=int, dest='batch_size', metavar='SIZE', default=200,
                            help="Split work to be done into batches, with the specified number of "
                            "records in each batch.  Or, set this to 0 (zero) to disable batching. "
                            "Implementation for this may vary somewhat between importers; default "
                            "batch size is 200 records.")

        # treat changes as warnings?
        parser.add_argument('--warnings', '-W', action='store_true',
                            help="Set this flag if you expect a \"clean\" import, and wish for any "
                            "changes which do occur to be processed further and/or specially.  The "
                            "behavior of this flag is ultimately up to the import handler, but the "
                            "default is to send an email notification.")

        # dry run?
        parser.add_argument('--dry-run', action='store_true',
                            help="Go through the full motions and allow logging etc. to "
                            "occur, but rollback (abort) the transaction at the end.")

    def run(self, args):
        log.info("begin `{} {}` for data models: {}".format(
                self.parent_name, self.name, ', '.join(args.models or ["(ALL)"])))

        handler = self.get_handler(args=args)
        models = args.models or handler.get_default_keys()
        log.debug("using handler: {}".format(handler))
        log.debug("importing models: {}".format(models))
        log.debug("args are: {}".format(args))

        kwargs = {
            'dry_run': args.dry_run,
            'warnings': args.warnings,
            'create': args.create,
            'max_create': args.max_create,
            'update': args.update,
            'max_update': args.max_update,
            'delete': args.delete,
            'max_delete': args.max_delete,
            'max_total': args.max_total,
            'progress': self.progress,
            'args': args,
        }
        handler.import_data(*models, **kwargs)

        # TODO: should this logging happen elsewhere / be customizable?
        if args.dry_run:
            log.info("dry run, so transaction was rolled back")
        else:
            log.info("transaction was committed")


class ExportRattail(ImportSubcommand):
    """
    Export data to another Rattail database
    """
    name = 'export-rattail'
    description = __doc__.strip()
    default_handler_spec = 'rattail.importing.rattail:FromRattailToRattailExport'
    default_dbkey = 'host'

    def get_handler_factory(self, **kwargs):
        spec = self.config.get('rattail.exporting', 'rattail.handler',
                               default=self.default_handler_spec)
        return load_object(spec)

    def add_parser_args(self, parser):
        super(ExportRattail, self).add_parser_args(parser)
        parser.add_argument('--dbkey', metavar='KEY', default=self.default_dbkey,
                            help="Config key for database engine to be used as the \"target\" "
                            "Rattail system, i.e. where data will be exported.  This key must "
                            "be defined in the [rattail.db] section of your config file.")

    def get_handler_kwargs(self, **kwargs):
        if 'args' in kwargs:
            kwargs['dbkey'] = kwargs['args'].dbkey
        return kwargs


class ImportToRattail(ImportSubcommand):
    """
    Generic base class for commands which import *to* a Rattail system.
    """
    # subclass must set these!
    handler_key = None
    default_handler_spec = None

    def get_handler_factory(self, **kwargs):
        spec = self.config.get('rattail.importing', '{}.handler'.format(self.handler_key),
                               default=self.default_handler_spec)
        return load_object(spec)


class ImportRattail(ImportToRattail):
    """
    Import data from another Rattail database
    """
    name = 'import-rattail'
    description = __doc__.strip()
    handler_key = 'rattail'
    default_handler_spec = 'rattail.importing.rattail:FromRattailToRattailImport'
    accepts_dbkey_param = True

    def add_parser_args(self, parser):
        super(ImportRattail, self).add_parser_args(parser)
        if self.accepts_dbkey_param:
            parser.add_argument('--dbkey', metavar='KEY', default='host',
                                help="Config key for database engine to be used as the Rattail "
                                "\"host\", i.e. the source of the data to be imported.  This key "
                                "must be defined in the [rattail.db] section of your config file.  "
                                "Defaults to 'host'.")

    def get_handler_kwargs(self, **kwargs):
        if self.accepts_dbkey_param:
            if 'args' in kwargs:
                kwargs['dbkey'] = kwargs['args'].dbkey
        return kwargs


class ImportRattailBulk(ImportRattail):
    """
    Bulk-import data from another Rattail database
    """
    name = 'import-rattail-bulk'
    description = __doc__.strip()
    handler_key = 'rattail_bulk'
    default_handler_spec = 'rattail.importing.rattail_bulk:BulkFromRattailToRattail'


class ImportSampleData(ImportToRattail):
    """
    Import sample data to a Rattail database
    """
    name = 'import-sample'
    description = __doc__.strip()
    handler_key = 'sample'
    default_handler_spec = 'rattail.importing.sample:FromSampleToRattail'


class ImportVersions(ImportRattail):
    """
    Make initial versioned records for data models
    """
    name = 'import-versions'
    description = __doc__.strip()
    handler_key = 'versions'
    default_handler_spec = 'rattail.importing.versions:FromRattailToRattailVersions'
    accepts_dbkey_param = False
    default_comment = "import catch-up versions"

    def add_parser_args(self, parser):
        super(ImportVersions, self).add_parser_args(parser)
        parser.add_argument('--comment', type=six.text_type, default=self.default_comment,
                            help="Comment to be recorded with the transaction.  "
                            "Default is \"{}\".".format(self.default_comment))

    def get_handler_kwargs(self, **kwargs):
        if 'args' in kwargs:
            kwargs['comment'] = kwargs['args'].comment
        return kwargs

    def run(self, args):
        if not self.config.versioning_enabled():
            self.stderr.write("Continuum versioning is not enabled, per config\n")
            sys.exit(1)
        super(ImportVersions, self).run(args)
